"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.assertTypeStructOrDefinedLinkNode = exports.isTypeStructOrDefinedLinkNode = exports.assertTypeNode = exports.isTypeNode = exports.createTypeNodeFromIdl = void 0;
const idl_1 = require("../idl");
const TypeArrayNode_1 = require("./TypeArrayNode");
const TypeBoolNode_1 = require("./TypeBoolNode");
const TypeBytesNode_1 = require("./TypeBytesNode");
const TypeDefinedLinkNode_1 = require("./TypeDefinedLinkNode");
const TypeEnumNode_1 = require("./TypeEnumNode");
const TypeMapNode_1 = require("./TypeMapNode");
const TypeNumberNode_1 = require("./TypeNumberNode");
const TypeOptionNode_1 = require("./TypeOptionNode");
const TypePublicKeyNode_1 = require("./TypePublicKeyNode");
const TypeSetNode_1 = require("./TypeSetNode");
const TypeStringNode_1 = require("./TypeStringNode");
const TypeStructNode_1 = require("./TypeStructNode");
const TypeTupleNode_1 = require("./TypeTupleNode");
const TYPE_NODE_CLASSES = [
    'TypeArrayNode',
    'TypeBoolNode',
    'TypeBytesNode',
    'TypeDefinedLinkNode',
    'TypeEnumNode',
    'TypeMapNode',
    'TypeNumberNode',
    'TypeNumberWrapperNode',
    'TypeOptionNode',
    'TypePublicKeyNode',
    'TypeSetNode',
    'TypeStringNode',
    'TypeStructNode',
    'TypeTupleNode',
];
function isArrayOfSize(array, size) {
    return Array.isArray(array) && array.length === size;
}
const createTypeNodeFromIdl = (idlType) => {
    // Leaf.
    if (typeof idlType === 'string' && idl_1.IDL_TYPE_LEAVES.includes(idlType)) {
        if (idlType === 'bool')
            return new TypeBoolNode_1.TypeBoolNode();
        if (idlType === 'string')
            return new TypeStringNode_1.TypeStringNode();
        if (idlType === 'publicKey')
            return new TypePublicKeyNode_1.TypePublicKeyNode();
        if (idlType === 'bytes')
            return new TypeBytesNode_1.TypeBytesNode();
        return new TypeNumberNode_1.TypeNumberNode(idlType);
    }
    // Ensure eveything else is an object.
    if (typeof idlType !== 'object') {
        throw new Error(`TypeNode: Unsupported type ${JSON.stringify(idlType)}`);
    }
    // Array.
    if ('array' in idlType && isArrayOfSize(idlType.array, 2)) {
        return TypeArrayNode_1.TypeArrayNode.fromIdl(idlType);
    }
    // Vec.
    if ('vec' in idlType) {
        return TypeArrayNode_1.TypeArrayNode.fromIdl(idlType);
    }
    // Defined link.
    if ('defined' in idlType && typeof idlType.defined === 'string') {
        return new TypeDefinedLinkNode_1.TypeDefinedLinkNode(idlType.defined);
    }
    // Enum.
    if ('kind' in idlType && idlType.kind === 'enum' && 'variants' in idlType) {
        return TypeEnumNode_1.TypeEnumNode.fromIdl(idlType);
    }
    // Map.
    if (('hashMap' in idlType && isArrayOfSize(idlType.hashMap, 2)) ||
        ('bTreeMap' in idlType && isArrayOfSize(idlType.bTreeMap, 2))) {
        return TypeMapNode_1.TypeMapNode.fromIdl(idlType);
    }
    // Option.
    if ('option' in idlType || 'coption' in idlType) {
        return TypeOptionNode_1.TypeOptionNode.fromIdl(idlType);
    }
    // Set.
    if ('hashSet' in idlType || 'bTreeSet' in idlType) {
        return TypeSetNode_1.TypeSetNode.fromIdl(idlType);
    }
    // Struct.
    if ('kind' in idlType && idlType.kind === 'struct') {
        return TypeStructNode_1.TypeStructNode.fromIdl(idlType);
    }
    // Tuple.
    if ('tuple' in idlType && Array.isArray(idlType.tuple)) {
        return TypeTupleNode_1.TypeTupleNode.fromIdl(idlType);
    }
    // Throw an error for unsupported types.
    throw new Error(`TypeNode: Unsupported type ${JSON.stringify(idlType)}`);
};
exports.createTypeNodeFromIdl = createTypeNodeFromIdl;
function isTypeNode(node) {
    return !!node && TYPE_NODE_CLASSES.includes(node.nodeClass);
}
exports.isTypeNode = isTypeNode;
function assertTypeNode(node) {
    var _a;
    if (!isTypeNode(node)) {
        throw new Error(`Expected TypeNode, got ${(_a = node === null || node === void 0 ? void 0 : node.nodeClass) !== null && _a !== void 0 ? _a : 'null'}.`);
    }
}
exports.assertTypeNode = assertTypeNode;
function isTypeStructOrDefinedLinkNode(node) {
    return (!!node &&
        (node.nodeClass === 'TypeStructNode' ||
            node.nodeClass === 'TypeDefinedLinkNode'));
}
exports.isTypeStructOrDefinedLinkNode = isTypeStructOrDefinedLinkNode;
function assertTypeStructOrDefinedLinkNode(node) {
    var _a;
    if (!isTypeStructOrDefinedLinkNode(node)) {
        throw new Error(`Expected TypeStructNode | TypeDefinedLinkNode, got ${(_a = node === null || node === void 0 ? void 0 : node.nodeClass) !== null && _a !== void 0 ? _a : 'null'}.`);
    }
}
exports.assertTypeStructOrDefinedLinkNode = assertTypeStructOrDefinedLinkNode;
//# sourceMappingURL=TypeNode.js.map