"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.renderJavaScriptValueNode = void 0;
const utils_1 = require("../../utils");
const JavaScriptImportMap_1 = require("./JavaScriptImportMap");
function renderJavaScriptValueNode(value) {
    var _a;
    const imports = new JavaScriptImportMap_1.JavaScriptImportMap();
    switch (value.__kind) {
        case 'list':
        case 'tuple':
            const list = value.values.map((v) => renderJavaScriptValueNode(v));
            return {
                imports: imports.mergeWith(...list.map((c) => c.imports)),
                render: `[${list.map((c) => c.render).join(', ')}]`,
            };
        case 'set':
            const set = value.values.map((v) => renderJavaScriptValueNode(v));
            return {
                imports: imports.mergeWith(...set.map((c) => c.imports)),
                render: `new Set([${set.map((c) => c.render).join(', ')}])`,
            };
        case 'map':
            const map = value.values.map(([k, v]) => {
                const mapKey = renderJavaScriptValueNode(k);
                const mapValue = renderJavaScriptValueNode(v);
                return {
                    imports: mapKey.imports.mergeWith(mapValue.imports),
                    render: `[${mapKey.render}, ${mapValue.render}]`,
                };
            });
            return {
                imports: imports.mergeWith(...map.map((c) => c.imports)),
                render: `new Map([${map.map((c) => c.render).join(', ')}])`,
            };
        case 'struct':
            const struct = Object.entries(value.values).map(([k, v]) => {
                const structValue = renderJavaScriptValueNode(v);
                return {
                    imports: structValue.imports,
                    render: `${k}: ${structValue.render}`,
                };
            });
            return {
                imports: imports.mergeWith(...struct.map((c) => c.imports)),
                render: `{ ${struct.map((c) => c.render).join(', ')} }`,
            };
        case 'enum':
            const enumName = (0, utils_1.pascalCase)(value.enumType);
            const variantName = (0, utils_1.pascalCase)(value.variant);
            const rawDependency = (_a = value.dependency) !== null && _a !== void 0 ? _a : 'generated';
            const dependency = rawDependency === 'generated' ? 'generatedTypes' : rawDependency;
            if (value.value === 'scalar') {
                return {
                    imports: imports.add(dependency, enumName),
                    render: `${enumName}.${variantName}`,
                };
            }
            const enumFn = (0, utils_1.camelCase)(value.enumType);
            imports.add(dependency, enumFn);
            if (value.value === 'empty') {
                return { imports, render: `${enumFn}('${variantName}')` };
            }
            const enumValue = renderJavaScriptValueNode(value.value);
            const fields = enumValue.render;
            imports.mergeWith(enumValue.imports);
            return {
                imports,
                render: `${enumFn}('${variantName}', ${fields})`,
            };
        case 'optionSome':
            const child = renderJavaScriptValueNode(value.value);
            return {
                imports: child.imports.add('core', 'some'),
                render: `some(${child.render})`,
            };
        case 'optionNone':
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().add('core', 'none'),
                render: 'none()',
            };
        case 'publicKey':
            return {
                imports: new JavaScriptImportMap_1.JavaScriptImportMap().add('core', 'publicKey'),
                render: `publicKey("${value.value}")`,
            };
        case 'string':
        case 'number':
        case 'boolean':
            return { imports, render: JSON.stringify(value.value) };
        default:
            const neverDefault = value;
            throw new Error(`Unexpected value type ${neverDefault.__kind}`);
    }
}
exports.renderJavaScriptValueNode = renderJavaScriptValueNode;
//# sourceMappingURL=RenderJavaScriptValueNode.js.map