"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetDefaultValidatorBagVisitor = void 0;
const nodes = __importStar(require("../../nodes"));
const NodeStack_1 = require("../NodeStack");
const ValidatorBag_1 = require("../ValidatorBag");
const GetResolvedInstructionAccountsVisitor_1 = require("./GetResolvedInstructionAccountsVisitor");
class GetDefaultValidatorBagVisitor {
    constructor() {
        this.stack = new NodeStack_1.NodeStack();
        this.program = null;
        this.definedTypes = new Set();
    }
    visitRoot(root) {
        // Register defined types to make sure links are valid.
        root.allDefinedTypes.forEach((type) => this.definedTypes.add(type.name));
        this.pushNode(root);
        const bags = root.programs.map((program) => program.accept(this));
        this.popNode();
        return new ValidatorBag_1.ValidatorBag().mergeWith(bags);
    }
    visitProgram(program) {
        this.pushNode(program);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!program.metadata.name) {
            bag.error('Program has no name.', program, this.stack);
        }
        if (!program.metadata.publicKey) {
            bag.error('Program has no public key.', program, this.stack);
        }
        if (!program.metadata.version) {
            bag.warn('Program has no version.', program, this.stack);
        }
        if (!program.metadata.origin) {
            bag.info('Program has no origin.', program, this.stack);
        }
        bag.mergeWith([
            ...program.accounts.map((node) => node.accept(this)),
            ...program.instructions.map((node) => node.accept(this)),
            ...program.definedTypes.map((node) => node.accept(this)),
            ...program.errors.map((node) => node.accept(this)),
        ]);
        this.popNode();
        return bag;
    }
    visitAccount(account) {
        this.pushNode(account);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!account.name) {
            bag.error('Account has no name.', account, this.stack);
        }
        bag.mergeWith([account.type.accept(this)]);
        this.popNode();
        return bag;
    }
    visitInstruction(instruction) {
        this.pushNode(instruction);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!instruction.name) {
            bag.error('Instruction has no name.', instruction, this.stack);
        }
        // Check for duplicate account names.
        const accountNameHistogram = new Map();
        instruction.accounts.forEach((account) => {
            var _a;
            if (!account.name) {
                bag.error('Instruction account has no name.', instruction, this.stack);
                return;
            }
            const count = ((_a = accountNameHistogram.get(account.name)) !== null && _a !== void 0 ? _a : 0) + 1;
            accountNameHistogram.set(account.name, count);
            // Only throw an error once per duplicated names.
            if (count === 2) {
                bag.error(`Account name "${account.name}" is not unique in instruction "${instruction.name}".`, instruction, this.stack);
            }
        });
        // Check for cyclic dependencies in account defaults.
        const cyclicCheckVisitor = new GetResolvedInstructionAccountsVisitor_1.GetResolvedInstructionAccountsVisitor();
        try {
            instruction.accept(cyclicCheckVisitor);
        }
        catch (error) {
            bag.error(cyclicCheckVisitor.getError(), instruction, this.stack);
        }
        bag.mergeWith([instruction.args.accept(this)]);
        // Check sub-instructions.
        bag.mergeWith(instruction.subInstructions.map((ix) => ix.accept(this)));
        this.popNode();
        return bag;
    }
    visitDefinedType(definedType) {
        this.pushNode(definedType);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!definedType.name) {
            bag.error('Defined type has no name.', definedType, this.stack);
        }
        bag.mergeWith([definedType.type.accept(this)]);
        this.popNode();
        return bag;
    }
    visitError(error) {
        this.pushNode(error);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!error.name) {
            bag.error('Error has no name.', error, this.stack);
        }
        if (typeof error.code !== 'number') {
            bag.error('Error has no code.', error, this.stack);
        }
        if (!error.message) {
            bag.warn('Error has no message.', error, this.stack);
        }
        this.popNode();
        return bag;
    }
    visitTypeArray(typeArray) {
        this.pushNode(typeArray);
        const bag = typeArray.item.accept(this);
        this.popNode();
        return bag;
    }
    visitTypeDefinedLink(typeDefinedLink) {
        this.pushNode(typeDefinedLink);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!typeDefinedLink.name) {
            bag.error('Pointing to a defined type with no name.', typeDefinedLink, this.stack);
        }
        else if (typeDefinedLink.dependency === 'generated' &&
            !this.definedTypes.has(typeDefinedLink.name)) {
            bag.error(`Pointing to a missing defined type named "${typeDefinedLink.name}"`, typeDefinedLink, this.stack);
        }
        this.popNode();
        return bag;
    }
    visitTypeEnum(typeEnum) {
        this.pushNode(typeEnum);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!typeEnum.name) {
            bag.info('Enum has no name.', typeEnum, this.stack);
        }
        if (typeEnum.variants.length === 0) {
            bag.warn('Enum has no variants.', typeEnum, this.stack);
        }
        typeEnum.variants.forEach((variant) => {
            if (!variant.name) {
                bag.error('Enum variant has no name.', typeEnum, this.stack);
            }
        });
        bag.mergeWith(typeEnum.variants.map((variant) => variant.accept(this)));
        this.popNode();
        return bag;
    }
    visitTypeEnumEmptyVariant(typeEnumEmptyVariant) {
        this.pushNode(typeEnumEmptyVariant);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!typeEnumEmptyVariant.name) {
            bag.error('Enum variant has no name.', typeEnumEmptyVariant, this.stack);
        }
        this.popNode();
        return bag;
    }
    visitTypeEnumStructVariant(typeEnumStructVariant) {
        this.pushNode(typeEnumStructVariant);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!typeEnumStructVariant.name) {
            bag.error('Enum variant has no name.', typeEnumStructVariant, this.stack);
        }
        bag.mergeWith([typeEnumStructVariant.struct.accept(this)]);
        this.popNode();
        return bag;
    }
    visitTypeEnumTupleVariant(typeEnumTupleVariant) {
        this.pushNode(typeEnumTupleVariant);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!typeEnumTupleVariant.name) {
            bag.error('Enum variant has no name.', typeEnumTupleVariant, this.stack);
        }
        bag.mergeWith([typeEnumTupleVariant.tuple.accept(this)]);
        this.popNode();
        return bag;
    }
    visitTypeMap(typeMap) {
        this.pushNode(typeMap);
        const bag = new ValidatorBag_1.ValidatorBag();
        bag.mergeWith([typeMap.key.accept(this), typeMap.value.accept(this)]);
        this.popNode();
        return bag;
    }
    visitTypeOption(typeOption) {
        this.pushNode(typeOption);
        const bag = typeOption.item.accept(this);
        this.popNode();
        return bag;
    }
    visitTypeSet(typeSet) {
        this.pushNode(typeSet);
        const bag = typeSet.item.accept(this);
        this.popNode();
        return bag;
    }
    visitTypeStruct(typeStruct) {
        this.pushNode(typeStruct);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!typeStruct.name) {
            bag.info('Struct has no name.', typeStruct, this.stack);
        }
        // Check for duplicate field names.
        const fieldNameHistogram = new Map();
        typeStruct.fields.forEach((field) => {
            var _a;
            if (!field.name)
                return; // Handled by TypeStructField
            const count = ((_a = fieldNameHistogram.get(field.name)) !== null && _a !== void 0 ? _a : 0) + 1;
            fieldNameHistogram.set(field.name, count);
            // Only throw an error once per duplicated names.
            if (count === 2) {
                bag.error(`Struct field name "${field.name}" is not unique.`, field, this.stack);
            }
        });
        bag.mergeWith(typeStruct.fields.map((field) => field.accept(this)));
        this.popNode();
        return bag;
    }
    visitTypeStructField(typeStructField) {
        this.pushNode(typeStructField);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (!typeStructField.name) {
            bag.error('Struct field has no name.', typeStructField, this.stack);
        }
        bag.mergeWith([typeStructField.type.accept(this)]);
        this.popNode();
        return bag;
    }
    visitTypeTuple(typeTuple) {
        this.pushNode(typeTuple);
        const bag = new ValidatorBag_1.ValidatorBag();
        if (typeTuple.items.length === 0) {
            bag.warn('Tuple has no items.', typeTuple, this.stack);
        }
        bag.mergeWith(typeTuple.items.map((node) => node.accept(this)));
        this.popNode();
        return bag;
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    visitTypeBool(typeBool) {
        return new ValidatorBag_1.ValidatorBag();
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    visitTypeBytes(typeBytes) {
        return new ValidatorBag_1.ValidatorBag();
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    visitTypeNumber(typeNumber) {
        return new ValidatorBag_1.ValidatorBag();
    }
    visitTypeNumberWrapper(typeNumberWrapper) {
        this.pushNode(typeNumberWrapper);
        const bag = typeNumberWrapper.item.accept(this);
        this.popNode();
        return bag;
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    visitTypePublicKey(typePublicKey) {
        return new ValidatorBag_1.ValidatorBag();
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    visitTypeString(typeString) {
        return new ValidatorBag_1.ValidatorBag();
    }
    pushNode(node) {
        this.stack.push(node);
        if (nodes.isProgramNode(node)) {
            this.program = node;
        }
    }
    popNode() {
        const node = this.stack.pop();
        if (node && nodes.isProgramNode(node)) {
            this.program = null;
        }
        return node;
    }
}
exports.GetDefaultValidatorBagVisitor = GetDefaultValidatorBagVisitor;
//# sourceMappingURL=GetDefaultValidatorBagVisitor.js.map