"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetNodeInlineStringVisitor = void 0;
const ROOT_PREFIX = 'R';
const PROGRAM_PREFIX = 'P';
const ACCOUNT_PREFIX = 'A';
const INSTRUCTION_PREFIX = 'I';
const TYPE_PREFIX = 'T';
const ERROR_PREFIX = 'E';
class GetNodeInlineStringVisitor {
    visitRoot(root) {
        const children = root.programs.map((program) => program.accept(this));
        return `${ROOT_PREFIX}(${children.join(',')})`;
    }
    visitProgram(program) {
        const children = [
            ...program.accounts.map((account) => account.accept(this)),
            ...program.instructionsWithSubs.map((ix) => ix.accept(this)),
            ...program.definedTypes.map((type) => type.accept(this)),
            ...program.errors.map((type) => type.accept(this)),
        ];
        return `${PROGRAM_PREFIX}[${program.name}](${children.join(',')})`;
    }
    visitAccount(account) {
        const child = account.type.accept(this);
        return `${ACCOUNT_PREFIX}[${account.name}](${child})`;
    }
    visitInstruction(instruction) {
        const accounts = instruction.accounts.map((account) => account.name);
        const args = instruction.args.accept(this);
        return (`${INSTRUCTION_PREFIX}[${instruction.name}](` +
            `accounts:(${accounts.join(',')}),` +
            `args:(${args})` +
            `)`);
    }
    visitDefinedType(definedType) {
        const child = definedType.type.accept(this);
        return `${TYPE_PREFIX}[${definedType.name}](${child})`;
    }
    visitError(error) {
        return `${ERROR_PREFIX}[${error.name}]`;
    }
    visitTypeArray(typeArray) {
        const item = typeArray.item.accept(this);
        const size = this.displayArrayLikeSize(typeArray.size);
        return `array(${item};${size})`;
    }
    visitTypeDefinedLink(typeDefinedLink) {
        return `link(${typeDefinedLink.name};${typeDefinedLink.dependency})`;
    }
    visitTypeEnum(typeEnum) {
        const children = typeEnum.variants.map((variant) => variant.accept(this));
        return `enum[${typeEnum.name}](${children.join(',')})`;
    }
    visitTypeEnumEmptyVariant(typeEnumEmptyVariant) {
        return typeEnumEmptyVariant.name;
    }
    visitTypeEnumStructVariant(typeEnumStructVariant) {
        const child = typeEnumStructVariant.struct.accept(this);
        return `${typeEnumStructVariant.name}:${child}`;
    }
    visitTypeEnumTupleVariant(typeEnumTupleVariant) {
        const child = typeEnumTupleVariant.tuple.accept(this);
        return `${typeEnumTupleVariant.name}:${child}`;
    }
    visitTypeMap(typeMap) {
        const key = typeMap.key.accept(this);
        const value = typeMap.value.accept(this);
        const size = this.displayArrayLikeSize(typeMap.size);
        return `map(${key},${value};${size})`;
    }
    visitTypeOption(typeOption) {
        const item = typeOption.item.accept(this);
        const prefix = typeOption.prefix.accept(this);
        const fixed = typeOption.fixed ? ';fixed' : '';
        return `option(${item};${prefix + fixed})`;
    }
    visitTypeSet(typeSet) {
        const item = typeSet.item.accept(this);
        const size = this.displayArrayLikeSize(typeSet.size);
        return `set(${item};${size})`;
    }
    visitTypeStruct(typeStruct) {
        const children = typeStruct.fields.map((field) => field.accept(this));
        return `struct[${typeStruct.name}](${children.join(',')})`;
    }
    visitTypeStructField(typeStructField) {
        const child = typeStructField.type.accept(this);
        return `${typeStructField.name}:${child}`;
    }
    visitTypeTuple(typeTuple) {
        const children = typeTuple.items.map((item) => item.accept(this));
        return `tuple(${children.join(',')})`;
    }
    visitTypeBool(typeBool) {
        return typeBool.toString();
    }
    visitTypeBytes(typeBytes) {
        return typeBytes.toString();
    }
    visitTypeNumber(typeNumber) {
        return typeNumber.toString();
    }
    visitTypeNumberWrapper(typeNumberWrapper) {
        const item = typeNumberWrapper.item.accept(this);
        const { wrapper } = typeNumberWrapper;
        switch (wrapper.kind) {
            case 'DateTime':
                return `DateTime(${item})`;
            case 'Amount':
                return `Amount(${item},${wrapper.identifier},${wrapper.decimals})`;
            case 'SolAmount':
                return `SolAmount(${item})`;
            default:
                return item;
        }
    }
    visitTypePublicKey() {
        return 'publicKey';
    }
    visitTypeString(typeString) {
        return typeString.toString();
    }
    displayArrayLikeSize(size) {
        if (size.kind === 'fixed')
            return `${size.size}`;
        if (size.kind === 'prefixed')
            return size.prefix.accept(this);
        return 'remainder';
    }
}
exports.GetNodeInlineStringVisitor = GetNodeInlineStringVisitor;
//# sourceMappingURL=GetNodeInlineStringVisitor.js.map